"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports._search = exports._sl = exports._copy = exports._make = exports._failIfNotExists = exports._exists = void 0;
const path_1 = require("path");
const fs_1 = require("fs");
const os_1 = require("os");
const _exists = (path) => __awaiter(void 0, void 0, void 0, function* () {
    return new Promise((_resolve, reject) => {
        fs_1.access(path, error => {
            if (error) {
                return _resolve(false);
            }
            return _resolve(true);
        });
    });
});
exports._exists = _exists;
const _failIfNotExists = (path) => __awaiter(void 0, void 0, void 0, function* () {
    const exists = yield exports._exists(path);
    if (!exists) {
        throw `${path} does not exist`;
    }
});
exports._failIfNotExists = _failIfNotExists;
/**
 * @param path to make recursively
 */
const _make = (path) => __awaiter(void 0, void 0, void 0, function* () {
    return new Promise((_resolve, reject) => {
        if (0 === path.length || '.' === path || '/' === path) {
            return _resolve();
        }
        exports._make(path_1.dirname(path)).then(() => exports._exists(path)).then(exists => {
            if (exists) {
                return _resolve();
            }
            fs_1.mkdir(path, error => {
                if (error) {
                    return reject(error);
                }
                return _resolve();
            });
        });
    });
});
exports._make = _make;
/**
 * NOTE that destination is overwritten if it already exists
 */
const _copy = (source, destination) => __awaiter(void 0, void 0, void 0, function* () {
    return new Promise((_resolve, reject) => {
        exports._make(path_1.dirname(destination)).then(() => {
            fs_1.copyFile(source, destination, error => {
                if (error) {
                    return reject(error);
                }
                return _resolve();
            });
        });
    });
});
exports._copy = _copy;
/**
 * NOTE that it won't fail in case target exist
 */
const _sl = (source, target, type, log) => __awaiter(void 0, void 0, void 0, function* () {
    return new Promise((_resolve, reject) => {
        // not sure what it does when path is not absolute path
        fs_1.symlink(path_1.resolve(source), path_1.resolve(target), type, error => {
            if (error) {
                if ('EEXIST' === error.code) {
                    if (log) {
                        log(`${target} already exists, EEXIST:${os_1.constants.errno.EEXIST}, error:${error.errno}`);
                    }
                    return _resolve();
                }
                return reject(error);
            }
            return _resolve();
        });
    });
});
exports._sl = _sl;
/**
 * search for a basename(@path) starting from dirname(@path) up to process.cwd()
 * @param path name to search for, e.g. package.json
 * @returns absolute path to existing basename(@path) or null in case it do not exists all the way up to process.cwd()
 */
const _search = (path) => __awaiter(void 0, void 0, void 0, function* () {
    const directory = path_1.resolve(path_1.dirname(path));
    // is it inside process.cwd()
    if (!directory.startsWith(process.cwd())) {
        return null;
    }
    const exists = yield exports._exists(path);
    if (exists) {
        // resolve as path has to be absolute
        return path_1.resolve(path);
    }
    return yield exports._search(path_1.join(path_1.dirname(directory), path_1.basename(path)));
});
exports._search = _search;
//# sourceMappingURL=fs.js.map