"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
const stream_1 = require("stream");
const child_process_1 = require("child_process");
const path_1 = require("path");
const globby = require("globby");
const fs_1 = require("./fs");
class ThePlugin {
    constructor(serverless, options) {
        this.serverless = serverless;
        this.options = options;
        this.hooks = {
            'before:package:createDeploymentArtifacts': this.build.bind(this),
            'after:package:createDeploymentArtifacts': this.dmp.bind(this)
        };
    }
    build() {
        return __awaiter(this, void 0, void 0, function* () {
            this.serverless.cli.log(`serverless.config.servicePath:${this.serverless.config.servicePath}`);
            this.serverless.cli.log(`serverless.service.package:${JSON.stringify(this.serverless.service.package)}`);
            this.serverless.service.package.include = [...this.serverless.service.package.include || [], path_1.join('package.json')];
            {
                // have to explicitly include everithing as we are adding ./** as exclude pattern for each single function
                this.serverless.cli.log('npm ls --only production --parseable --silent');
                // have to execute it in service package root (where serverless.[yml|json|js] is, and where serverless command is executed) as in .build, node_modules is symlinked and npm list command cannot work on symlinked directory
                const deps = yield dependencies(process.cwd());
                const depsar = deps.map(path => path_1.relative(process.cwd(), path)).filter(path => path.length !== 0);
                for (const path of depsar) {
                    const pathRelativeToOriginalServicePath = `${path}/**`;
                    this.serverless.cli.log(`will push ${pathRelativeToOriginalServicePath} to directories which should be included in the deployment package`);
                    this.serverless.service.package.include.push(pathRelativeToOriginalServicePath);
                }
            }
            // transpile projects to emitte files
            const cache = new Map();
            const functionNames = this.serverless.service.getAllFunctions();
            const acb = (name) => __awaiter(this, void 0, void 0, function* () {
                const _function = this.serverless.service.functions[name];
                this.serverless.cli.log(`serverless.service.functions[${name}].handler:${_function.handler}`);
                // looks like there is undocumented property "disable"
                this.serverless.cli.log(`serverless.service.functions[${name}].package:${JSON.stringify(_function.package)}`);
                // skip this function
                if (!_function.package.individually) {
                    this.serverless.cli.log(`skipping serverless.service.functions[${name}] as package individually is not set to true`);
                    return;
                }
                // where handler's .ts is... relative to service package root (where serverless.[yml|json|js] is, and where serverless command is executed)
                const pathToHandlerRelativeToServicePackageRoot = _function.handler.replace(path_1.extname(_function.handler), '.ts');
                this.serverless.cli.log(`path to handler relative to service package root:${pathToHandlerRelativeToServicePackageRoot}`);
                // check that ....
                yield fs_1._failIfNotExists(path_1.join(process.cwd(), pathToHandlerRelativeToServicePackageRoot));
                // absolute path to function's package.json
                const absolutePathToFunctionPackageJSON = yield fs_1._search(path_1.join(path_1.dirname(pathToHandlerRelativeToServicePackageRoot), 'package.json'));
                if (!absolutePathToFunctionPackageJSON) {
                    // NOTE that will make function to return rejected promise
                    throw `can not find package.json all the way up to ${process.cwd()}`;
                }
                // relative to service package root
                const relativePathToFunctionPackageJSON = path_1.relative('.', absolutePathToFunctionPackageJSON);
                // absolute path to function's tsconfig.json
                const absolutePathToFunctionTSConfigJSON = yield fs_1._search(path_1.join(path_1.dirname(pathToHandlerRelativeToServicePackageRoot), 'tsconfig.json'));
                if (!absolutePathToFunctionTSConfigJSON) {
                    // NOTE that will make function to return rejected promise
                    throw `can not find tsconfig.json all the way up to ${process.cwd()}`;
                }
                // copy any "extras" from the "include" section
                if (_function.package.include) {
                    const paths = yield globby(_function.package.include);
                    for (const path of paths) {
                        const source = path_1.resolve(path);
                        const destination = path_1.resolve(path_1.join(this.serverless.config.servicePath, path));
                        yield fs_1._copy(source, destination);
                    }
                }
                // relative to service package root
                const relativePathToFunctionTSConfigJSON = path_1.relative('.', absolutePathToFunctionTSConfigJSON);
                const hasBeenAlreadyTranspiled = Array.isArray(cache.get(relativePathToFunctionTSConfigJSON));
                this.serverless.cli.log(`tsc --listEmittedFiles --rootDir . --outDir .build --project ${relativePathToFunctionTSConfigJSON}`);
                // NOTE have to pass 1st argument (rootDir) as absolute path otherwise TCS will fail with "Paths must either both be absolute or both be relative"
                const files = hasBeenAlreadyTranspiled ? cache.get(relativePathToFunctionTSConfigJSON) : (yield tsc(process.cwd(), '.build', relativePathToFunctionTSConfigJSON)).filter(file => !file.endsWith('.tsbuildinfo'));
                if (!hasBeenAlreadyTranspiled) {
                    cache.set(relativePathToFunctionTSConfigJSON, files);
                }
                for (const file of files) {
                    const pathRelativeToServicePath = path_1.relative(this.serverless.config.servicePath, file);
                    this.serverless.cli.log(`will push ${pathRelativeToServicePath} to files which should be included in the deployment package for function: ${name}`);
                    _function.package.include = Array.isArray(_function.package.include) ? _function.package.include : [];
                    _function.package.include.push(pathRelativeToServicePath);
                }
                // in case function has its own package.json... few more things has to happen
                if ('package.json' !== relativePathToFunctionPackageJSON) {
                    // TODO it might fail in case target path sergment do not exists
                    yield fs_1._sl(relativePathToFunctionPackageJSON, path_1.join(this.serverless.config.servicePath, relativePathToFunctionPackageJSON), 'file', message => { this.serverless.cli.log(message); });
                    this.serverless.cli.log(`will push ${relativePathToFunctionPackageJSON} to files which should be included in the deployment package for function: ${name}`);
                    _function.package.include = [..._function.package.include || [], relativePathToFunctionPackageJSON];
                    const relativePathToNodeModules = path_1.join(path_1.dirname(relativePathToFunctionPackageJSON), 'node_modules');
                    const exists = yield fs_1._exists(relativePathToNodeModules);
                    if (exists) {
                        // TODO it might fail in case target path sergment do not exists
                        yield fs_1._sl(relativePathToNodeModules, path_1.join(this.serverless.config.servicePath, relativePathToNodeModules), 'dir', message => { this.serverless.cli.log(message); });
                        const relativePathToFunctionPackageRoot = path_1.dirname(relativePathToNodeModules);
                        const absolutePathToFunctionPackageRoot = path_1.resolve(relativePathToFunctionPackageRoot);
                        this.serverless.cli.log('npm ls --only production --parseable --silent');
                        const deps = yield dependencies(absolutePathToFunctionPackageRoot);
                        const depsar = deps.map(path => path_1.relative(absolutePathToFunctionPackageRoot, path)).filter(path => path.length !== 0);
                        for (const path of depsar) {
                            const pathRelativeToServicePackageRoot = path_1.join(relativePathToFunctionPackageRoot, path, '**');
                            this.serverless.cli.log(`will push ${pathRelativeToServicePackageRoot} to directories which should be included in the deployment package for function: ${name}`);
                            _function.package.include.push(pathRelativeToServicePackageRoot);
                        }
                    }
                }
                // have to exclude everything as otherwise, not sure why, almost everything gets inside package
                const patternToExcludeEverythingInsideServicePackage = './**';
                this.serverless.cli.log(`will push ${patternToExcludeEverythingInsideServicePackage} to directories which should be excluded in the deployment package for function: ${name}`);
                _function.package.exclude = [..._function.package.exclude || [], patternToExcludeEverythingInsideServicePackage];
            });
            for (const name of functionNames) {
                yield acb(name);
            }
        });
    }
    dmp() {
        this.serverless.cli.log(`this.serverless.config.servicePath:${this.serverless.config.servicePath}`);
        this.serverless.cli.log(`this.serverless.service.package:${JSON.stringify(this.serverless.service.package)}`);
        const functionNames = this.serverless.service.getAllFunctions();
        functionNames.forEach((name) => {
            const _function = this.serverless.service.functions[name];
            this.serverless.cli.log(`this.serverless.service.functions[${name}].handler:${_function.handler}`);
            this.serverless.cli.log(`this.serverless.service.functions[${name}].package:${JSON.stringify(_function.package)}`);
        });
    }
}
/**
 * tsc --listEmittedFiles --rootDir input --outDir output --project project
 * @param input Specifies the root directory of input files.
 * @param output Redirect output structure to the directory.
 * @param project path to directory containing a tsconfig.json file, or a path to a valid .json file containing the configurations.
 * @param cwd defaults to process.cwd()
 */
function tsc(input, output, project, cwd = process.cwd()) {
    return __awaiter(this, void 0, void 0, function* () {
        const chunks = [];
        // Buffer.con
        const writable = new stream_1.Writable({
            write: function (chunk, encoding, callback) {
                if (Buffer.isBuffer(chunk)) {
                    chunks.push(Buffer.isEncoding(encoding) ? chunk.toString(encoding) : chunk.toString());
                }
                else {
                    chunks.push(chunk);
                }
                callback();
                return true;
            }
        });
        try {
            // tsc --listEmittedFiles --rootDir input --outDir output --project project
            // TODO adopt tsc --build but keep in mind that --rootDir, --outDir and --project cannot be used togther with --build
            yield spawnWithPromise('tsc', ['--listEmittedFiles', '--rootDir', input, '--outDir', output, '--project', project], { cwd: cwd, stdio: [process.stdin, 'pipe', process.stderr] }, writable);
            // TSFILE: /absolute/path/to/emitted/file.[js|js.map|d.ts||d.ts.map]
            return chunks.reduce((result, chunk, index) => result + chunk, "").split('\n').filter(line => line.length !== 0 && line.startsWith('TSFILE:')).map(line => line.replace('TSFILE:', '').trim());
        }
        catch (error) {
            const message = chunks.reduce((result, chunk, index) => result + chunk, "");
            console.log(message);
            throw error;
        }
    });
}
function dependencies(cwd = process.cwd()) {
    return __awaiter(this, void 0, void 0, function* () {
        const chunks = [];
        // Buffer.con
        const writable = new stream_1.Writable({
            write: function (chunk, encoding, callback) {
                if (Buffer.isBuffer(chunk)) {
                    chunks.push(Buffer.isEncoding(encoding) ? chunk.toString(encoding) : chunk.toString());
                }
                else {
                    chunks.push(chunk);
                }
                callback();
                return true;
            }
        });
        try {
            // npm ls --only production --parseable --silent
            yield spawnWithPromise('npm', ['ls', '--only', 'production', '--parseable', '--silent'], { cwd: cwd, stdio: [process.stdin, 'pipe', process.stderr] }, writable);
            return chunks.reduce((result, chunk, index) => result + chunk, "").split('\n').filter(path => path.length !== 0);
        }
        catch (error) {
            const message = chunks.reduce((result, chunk, index) => result + chunk, "");
            console.log(message);
            throw error;
        }
    });
}
function spawnWithPromise(command, args, options, writable) {
    return new Promise((_resolve, _reject) => {
        const child = child_process_1.spawn(command, args !== null && args !== void 0 ? args : [], options !== null && options !== void 0 ? options : { stdio: 'pipe' });
        if (writable && child.stdout) {
            child.stdout.pipe(writable);
        }
        child.on('error', function (error) {
            console.error(error);
        });
        child.on('exit', function (code, signal) {
            if (code) {
                // TODO what if there is sensitive data that we don't what to show!
                const messsage = `${command + (Array.isArray(args) ? ' ' + args.join(' ') : '')} exited with code ${code} and signal ${signal}`;
                _reject(new Error(messsage));
            }
            _resolve();
        });
    });
}
module.exports = ThePlugin;
// module.exports = ThePlugin;
//# sourceMappingURL=index.js.map